#!/usr/bin/perl

#$Revision: 1528 $

# Copyright 2007, Fonality Inc. ALL RIGHTS RESERVED.
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301, USA.
# Author: Adrian Otto <support@fonality.com>
# Description: Fonality utility for PSTN card detection and configuration

use strict;
use Data::Dumper;
use Getopt::Long;
use File::Copy;
use LWP::Simple;
use Sys::Hostname;
use Carp;

$|++;
my($DEBUG) = 0;
my($VERBOSE) = 0;

main_sub();

##############################################################################
# main_sub: main program logic
##############################################################################
sub main_sub {
	my($WANONLY) = 0;
        my($RAW) = 0;
        my($CONFIG) = 0;
        my($UNLOAD_ONLY) = 0;
        my($LOAD_ONLY) = 0;
        my($QUIET) = 0;
        my($SUPPORTED) = 0;
        my($HELP) = 0;
        my($PROBE) = 0;

        my($getopt_result) = GetOptions (
                "wo" => \$WANONLY,
                "raw" => \$RAW,
                "c"  => \$CONFIG,
                "d"  => \$DEBUG,
                "u"  => \$UNLOAD_ONLY,
                "l"  => \$LOAD_ONLY,
                "q"  => \$QUIET,
                "v"  => \$VERBOSE,
                "s"  => \$SUPPORTED,
                "h"  => \$HELP,
                "p"  => \$PROBE,
        );

	if($QUIET && $VERBOSE) {
		print "\nERROR: You need to make up your mind if you want to be quiet or verbose.\n";
		show_usage();
	}

	if($QUIET && $DEBUG) {
		print "\nERROR: You need to make up your mind if you want to be quiet or debug.\n";
		show_usage();
	}

	if($HELP) { # -h
		show_usage();
	}

	if($SUPPORTED) { # -s
		show_supported_cards();
		exit(0);
	}

	if($PROBE) { # -p
		for my $i (25..28) {
			print "Probing channel $i\n";
			print Dumper(get_port_details($i));
		}
		exit(0);
	}

	if($UNLOAD_ONLY) { # -u
		unload_all_modules();
		exit(0);
	}

        my($cards) = find_pstn_cards();

	# clear out old config if there are no cards in system
	unless (scalar(@{$cards}) > 0) {
	    prune_old_config('/etc/zaptel.conf');	    
	    prune_old_config('/etc/asterisk/zapata.conf');
	}
	
	if($WANONLY) { # --wo configure only wanpipe and do not write zap*.conf

		config_wanpipe($WANONLY);
		force_minimal_zaptel_config();

		foreach my $card (@{$cards}) { 
		###  configure faxing on Sangoma A200 cards
			if($card->{'model'} eq 'A200') {
                    		configure_fax();
                	}
		}
	}

        if($RAW) { # --raw
		my($num_cards) = scalar(@{$cards});
		vprint("Displaying $num_cards cards");
                Data::Dumper->Purity(1);
                print Dumper($cards);
		exit(0);
	}
	if($CONFIG || $LOAD_ONLY) { # -c and/or -l
		unload_all_modules();
		configure_cards($cards,$LOAD_ONLY);
	
	}

	if($QUIET == 0) {
		my($num_cards) = scalar(@{$cards});
		if($num_cards == 1) {
			vprint("Found $num_cards card:");
		}
		elsif($num_cards > 1) {
			vprint("Found $num_cards cards:");
		}
		foreach my $card (@{$cards}) {
			print ">>> " if($VERBOSE);
			print $card->{'description'} . "\n";
		}
		if($num_cards == 0) {
			print ">>> " if($VERBOSE);
			print "No PSTN cards found\n";
		}
	}
}

##############################################################################
# show_usage: Print a usage message and exit.
##############################################################################
sub show_usage {
	print "\nUSAGE: $0 [options]\n\n";
	print "Options:\n\n";
	print "\t-l = Scan for cards, and load associated modules.\n";
	print "\t-u = Unload modules\n";
	print "\t-c = Generate /etc/zaptel.conf and /etc/asterisk/zapata.conf\n";
	print "\t-q = Do not print messages to terminal about cards found\n";
	print "\t-v = Print verbose status messages\n";
	print "\t-d = Print verbose status and debug messages to screen\n";
	print "\t-s = Print a list of supported cards, and exit\n\n";
	#print "\t--raw = Dump results in perl format to be read by eval()\n\n";
	print "\tWithout any options, prints a summary of cards to the terminal.\n\n";
	exit(0);
}

##############################################################################
# find_pstn_cards: Scan the system bus for PSTN interface cards
##############################################################################
sub find_pstn_cards {
	my($card_info) = get_card_info();
	my(@cards_found);
	open(LSPCI, "/sbin/lspci -vn |") || die $!;
	while(<LSPCI>) {
		foreach my $key (keys(%{$card_info})) {
			if(/$key/) {
				dprint("Adding " . $card_info->{$key}->{'description'} . " to list of cards found");
				push(@cards_found, $card_info->{$key});
			}
		}
	}
	close(LSPCI);

	return(\@cards_found);
}

##############################################################################
# configure_cards: Configure zaptel.conf and zapata.conf for the cards found
##############################################################################
sub configure_cards {
	my($cards) = shift(@_);
	my($LOAD_ONLY) = shift(@_);
	my($found_sangoma) = 0;
	my($found_digital) = 0;
	my($found_analog) = 0;
	my(%digital_drivers);
	my(%analog_drivers);
	my($spans,$channels,$new_chans,$new_spans);
	my $configure_fax = 0;
	my($loop_tmp);

	foreach my $card (@{$cards}) {
		$found_sangoma++ if($card->{'driver'} eq 'wanpipe');
		if($card->{'type'} eq 'digital') {
			$found_digital++;
			$digital_drivers{$card->{'driver'}}++;
		}
		elsif($card->{'type'} eq 'analog') {
			$found_analog++;
			$analog_drivers{$card->{'driver'}}++;
		}
		
		# configure faxing on Sangoma A200 cards
		if($card->{'model'} eq 'A200')
		{
		    $configure_fax = 1;
		}
	}

	if($LOAD_ONLY == 0) {
		# Check for FXS ports, and bail out if we have any
		my($f1) = '/etc/asterisk/fonality/external.conf';
		my($f2) = '/etc/asterisk/fonality/internal.conf';
		my($found_analog_ext) = 0;
		foreach my $f ($f1, $f2) {
			open(FILE, $f) || warn $!;
			while(<FILE>) {
				if(/ (\d+),hint,(Zap\/\d+)/i && 0) {
					$found_analog_ext++;
					print "

Found an analog extension $1 using port ${2}. 

This extension must be deleted before you can detect new cards. Please delete
this, and all other analog phone extensions and re-run this program.

Aborting after 5 seconds.\n\n";
					sleep(5);
					last;
				}
			}
			close(FILE);
			if($found_analog_ext) {
				return(0);
			}
		}
	}

	if($found_sangoma && ($LOAD_ONLY == 0)) {
		vprint("Configuring Sangoma card(s)...");
		stop_wanrouter();

		# Whack prior configuration and start over
		prune_conf_entry(
			'/etc/zaptel.conf',
			"# Sangoma Configuration START",
			"# Sangoma Configuration END",
		);
		prune_conf_entry(
			'/etc/asterisk/zapata.conf',
			"; Sangoma Configuration START",
			"; Sangoma Configuration END",
		);
		prune_single_line(
			'/etc/asterisk/zapata.conf',
			"callerid=asreceived"
		);

		# Configure the Sangoma wanpipe configuration files
		config_wanpipe();

		open(ZAP, "</etc/zaptel.conf") || warn $!;
		my @zap = <ZAP>;
		close(ZAP);

		open(ZAP, ">/etc/zaptel.conf");
		foreach (@zap)
		{
			s/^span=1,0,/span=1,1,/;
			print ZAP $_;
		}
		close(ZAP);

		open(AUTO, '/etc/asterisk/zapata-auto.conf') || warn $!;
		open(CONF, '>/etc/asterisk/zapata.conf') || warn $!;
		print CONF << "EOF";
[channels]
context=incoming
switchtype=national
signalling=pri_cpe
rxwink=300              ; Atlas seems to use long (250ms) winks
usecallerid=yes
hidecallerid=no
;callwaiting=yes
usecallingpres=yes
callwaitingcallerid=yes
;threewaycalling=yes
transfer=yes
cancallforward=yes
callreturn=yes
echocancel=yes
echocancelwhenbridged=yes
echotraining=yes
rxgain=0
txgain=0
group=2
callgroup=1
pickupgroup=1
immediate=no

EOF

		while(<AUTO>) {
			if(/; Sangoma Configuration/) {
				print CONF $_;
			}
			elsif(/;Sangoma/) {
				print CONF $_;
			}
			elsif(/^;/) {
				# Skip this line
			}
			else {
				print CONF $_;
			}
		}
		close(CONF);
		close(AUTO);
	}
	
	if($found_digital) {
		vprint("Configuring T1/E1 card(s)...");
		foreach my $name (keys(%digital_drivers)) {
			my $preloaded = module_is_loaded($name);
			if(load_module($name)) {
				($spans,$channels,$new_chans,$new_spans) = check_for_chans($spans,$channels);
				#dprint(Dumper($spans));
				#dprint(Dumper($channels));
				if(!$preloaded && $new_chans < 1) {
					dprint("When scanning for digital cards: No channels added by module: $name");
					unload_module($name);
					next;
				}
			}
		}
	}
	if($found_analog) {
		vprint("Configuring analog card(s)...");
		# flush messages
		system("/usr/sbin/logrotate -f /etc/logrotate.conf");
		foreach my $name (keys(%analog_drivers)) {
			dprint("Is $name preloaded?");
			my $preloaded = module_is_loaded($name);
			dprint("Loading $name");
			if(load_module($name)){ 
				($spans,$channels,$new_chans,$new_spans) = check_for_chans($spans,$channels);
				dprint("Checked for channels on $spans");
				#dprint(Dumper($spans));
				#dprint(Dumper($channels));
				if(!$preloaded && $new_chans < 1) {
					dprint("When scanning for analog cards: No channels added by module: $name");
					unload_module($name);
				}
			}
		}
	}

	if($found_analog || $found_sangoma) {
		dprint("Checking cable");
		if(power_cable_unplugged()) {
			print "FATAL ERROR: You must connect power cables to all cards with plugs\n\n";
			sleep(10);
			exit(1);
		}
	}

	if (
                (
                        ref($cards) eq "ARRAY" &&
                        scalar(@{$cards}) < 1
                ) ||
                (
                        ref($channels) eq "HASH" &&
                        scalar(keys(%{$channels})) < 1
                ) ||
                !ref($cards) ||
                !ref($channels)
	)
	{
		dprint("Loading ztdummy because no cards/channels found");
		load_module('ztdummy');
	}

	if($found_sangoma || $LOAD_ONLY) {
		dprint("Starting wanrouter");
		start_wanrouter();
		if ($LOAD_ONLY) { 
			dprint("Done loading modules. Exiting gracefully");
			if(ztcfg_fails()) {
				vprint("WARNING: Unable to configure PSTN card channels (ztcfg failure)");	
			}
			return(0);
		}
	}

	# POINT OF NO RETURN: We start setting up configuration files here
	if(scalar(keys(%{$channels}))) { # There is something to actually configure
		# Dump the channel parts of the old config (digium cards)
		unless($found_sangoma) {
			prune_old_config('/etc/zaptel.conf');
			prune_old_config('/etc/asterisk/zapata.conf');
		}

		foreach my $s (sort { $a <=> $b } keys(%{$spans})) {
			if($spans->{$s} =~ /wanpipe/) {
				next;
			}
			elsif(($spans->{$s} =~ /^WCTDM/i) ||
				($spans->{$s} =~ /^WCFXO/i)) {
				my(@chan_list) = sort { $a <=> $b } keys(%{get_channel_list_by_span($s)});
				# Pull out any prior configs for specific channels
				prune_channel_list(\@chan_list);

				# Configure the ports on the card
				if(configure_by_probe(\@chan_list)) {
					vprint("Span $s configured using auto-probe");
				}
				else {
					configure_by_guess(\@chan_list);
					vprint("Span $s configured by guessing");
				}
			}
			elsif($spans->{$s} =~ /^TE\d/i) {
				my($first_bchan,$last_bchan,$dchan) = get_chan_details_by_span($s);
				next unless($first_bchan && $last_bchan && $dchan);
				my($timing) = 1;
				$timing = 0 if($s > 1);
				update_conf_file(
					'/etc/zaptel.conf',
					"# Span " . $spans->{$s} . " START",
					"span=$s,$timing,0,esf,b8zs",
					"bchan=${first_bchan}-${last_bchan}",
					"dchan=${dchan}",
					"# Span " . $spans->{$s} . " END",
				);
				update_conf_file(
					'/etc/asterisk/zapata.conf',
					"; Span " . $spans->{$s} . " START",
					'switchtype = national',
					'signalling = pri_cpe',
					'context=incoming',
					'group = 3',
					"channel => ${first_bchan}-${last_bchan}",
					"; Span " . $spans->{$s} . " END",
				);

				# Pull out any prior configs for specific channels
				foreach my $c ($first_bchan..$dchan) {
					prune_channel($c);
				}
				foreach my $c ($first_bchan..$last_bchan) {
					prune_channel($c);
				}
			}
			elsif($spans->{$s} =~ /^WCT1/i) { # TE120 card
				my($first_bchan,$last_bchan,$dchan) = get_chan_details_by_span($s);
				next unless($first_bchan && $last_bchan && $dchan);
				my($timing) = 1;
				$timing = 0 if($s > 1);
				update_conf_file(
					'/etc/zaptel.conf',
					"# Span " . $spans->{$s} . " START",
					"span=$s,$timing,0,ccs,hdb3,crc4",
					"bchan=${first_bchan}-${last_bchan}",
					"dchan=${dchan}",
					"# Span " . $spans->{$s} . " END",
				);
				update_conf_file(
					'/etc/asterisk/zapata.conf',
					"# Span " . $spans->{$s} . " START",
					'switchtype = national',
					'signalling = pri_cpe',
					'context=incoming',
					'group = 1',
					"channel => ${first_bchan}-${last_bchan}",
					"# Span " . $spans->{$s} . " END",
				);

				# Pull out any prior configs for specific channels
				foreach my $c ($first_bchan..$dchan) {
					prune_channel($c);
				}
				foreach my $c ($first_bchan..$last_bchan) {
					prune_channel($c);
				    }
			    }
		    }
}

	if(ztcfg_fails()) {
		vprint("WARNING: Unable to configure PSTN card channels (ztcfg failure)");	
	}

	force_minimal_zaptel_config();
	
	#clean up zapata.conf and woomera.conf context entries
	system("/bin/sed -i -e 's/from-zaptel/incoming/' /etc/asterisk/{woomera,zapata}.conf");
	system("/bin/sed -i -e 's/from-internal/internal/' /etc/asterisk/{woomera,zapata}.conf");
	system("/bin/sed -i -e 's/from-pstn/incoming/' /etc/asterisk/{woomera,zapata}.conf");

	sync_conf_file('zapata.conf');
	sync_conf_file('zaptel.conf');

	if ($configure_fax)
	{
	    configure_fax();
	}
	
	if (-e "/etc/wanpipe/smg_bri.conf") {sync_conf_file('smg_bri.conf');}
	foreach $loop_tmp (</etc/wanpipe/wanpipe*.conf>) {
		(my($file_to_sync)) = ($loop_tmp =~ /.*(wanpipe\d+.conf)/i);
		sync_conf_file($file_to_sync);
	}

	configure_woomera_module();
	## better safe than sorry :P
	system("/sbin/chkconfig wanrouter off > /dev/null 2>&1");

}


##############################################################################
# configure_fax: configure card for faxing capability
##############################################################################
sub configure_fax
{
    stop_asterisk();
    stop_wanrouter();

    # figure out what wanpipe config (N) we need to deal with
    my $n;
    my $temp_hwprobe = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);
    lsystem("wanrouter hwprobe verbose > $temp_hwprobe");

    # read in hwprobe
    open HW, $temp_hwprobe || warn $!;
    while (<HW>)
    {
	if (/(\d+)\s\..*A200-SH\s:/)
	{
	    $n = $1;
	    last;
	}
    }
    close HW;
    unlink($temp_hwprobe);

    # add 'RM_NETWORK_SYNC = YES' to wanpipeN.conf by default  
    my $wanfile = "/etc/wanpipe/wanpipe$n.conf";
    my($temp_file) = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);

    open TEMP, ">$temp_file" || warn $!;
    open WAN, $wanfile || warn $!;
    while(<WAN>)
    {
	if (/RM_NETWORK_SYNC/)
	{
	    next;
	}
	elsif (/\[wanpipe$n\]/)
	{
	    print TEMP "$_";
	    print TEMP "RM_NETWORK_SYNC = YES\n";
	}
	else
	{
	    print TEMP "$_";
	}
    }
    close TEMP;
    close OUT;
    
    copy($temp_file, $wanfile) || warn $!;

    # start wanrouter with new config and see what it says
    my $wr_err = 0;
    my $wr_out = '/tmp/wr.out';
    lsystem("/usr/sbin/wanrouter start > /dev/null 2>$wr_out");

    open IN, $wr_out;
    while (<IN>) {
	if(/w${n}g1: unknown interface: No such device/i)
	{
	    $wr_err = 1;
	    last;
	}
    }
    close IN;
    unlink ($wr_out);

    if (!$wr_err)
    {
	# wanpipe loaded, yay
	return 1;
    }
    else
    {
	# if wanpipe fails to load, change it to 'NO' & restart

	stop_asterisk();
	stop_wanrouter();
	
	$temp_file = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);

	open TEMP, ">$temp_file" || warn $!;
	open WAN, $wanfile || warn $!;
	while(<WAN>)
	{
	    if (/RM_NETWORK_SYNC/)
	    {
		print TEMP "RM_NETWORK_SYNC = NO\n";
	    }
	    else
	    {
		print TEMP "$_";
	    }
	}
	close WAN;
	close TEMP;

	copy($temp_file, $wanfile) || warn $!;

	# start wanrouter
	start_wanrouter();
   }
}


##############################################################################
# configure_by_probe: Configure by probing the card (requires new drivers)
##############################################################################
sub configure_by_probe {
	my($chan_list) = shift(@_);

	unless(zaptel_version_supports_autoconfig()) {
		dprint("Zaptel version does not support auto-probe");
		return(0);
	}

	foreach my $c (@{$chan_list}) {
		my($port) = shift(@{get_port_details($c)}); # Probe the card for info
		if($port->{'porttype'} eq 'FXS') {
			configure_as_fxs_port($c);
		}
		elsif($port->{'porttype'} eq 'FXO') {
			configure_as_fxo_port($c);
		}
		if(ztcfg_fails()) {
			dprint("Checking ztcfg for channel $c failed, so removing config");
			prune_channel($c);
		}
	}

	return(1);
}

##############################################################################
# prune_channel_list
##############################################################################
sub prune_channel_list {
	my($chan_list) = shift(@_);
	foreach my $c (@{$chan_list}) {
		prune_channel($c);
	}
}

##############################################################################
# prune_channel: Prune configuration for a channel from zaptel and zapata
##############################################################################
sub prune_channel {
	my($c) = shift(@_);

	prune_conf_entry(
		'/etc/zaptel.conf',
		"# Channel $c START",
		"# Channel $c END",
	);
	prune_conf_entry(
		'/etc/asterisk/zapata.conf',
		"; Channel $c START",
		"; Channel $c END",
	);
	return(1);
}

##############################################################################
# configure_by_guess: Configure by guessing (legacy mode)
##############################################################################
sub configure_by_guess {

    # read channel info from /var/log/messages
    my $var_messages = '/var/log/messages';
    my @module_arr;
    open MSG, "<$var_messages";
    while (<MSG>) {

	if (/Module\s(\d+):\s(.*)$/) {
	    push(@module_arr, "$2");
	}
	elsif (/Port\s(\d+):.*(FXO|FXS)/)
	{
	    push (@module_arr, "$2");
	}
	elsif (/(FXO)\sPCI\sMaster\sabort/)
	{
	    push (@module_arr, "$1");
	}
    }
    for (my $c = 0; $c <= $#module_arr; $c++) {
	my $mod = $module_arr[$c];
	my $channel = $c+1;  # module starts at 0, but channel at 1
	if ($mod =~ /(FXO|FXS)/) {
	    my $fx = 'fxo';
	    my $context = 'internal';
	    if ($1 eq 'FXO') {
		$fx = 'fxs';
		    $context="incoming\n".
			"group = 0";
	    }
	    update_conf_file(
			     '/etc/zaptel.conf',
			     "# Channel $channel START",
			     $fx."ks=$channel",
			     "# Channel $channel END",
			     );
	    update_conf_file(
			     '/etc/asterisk/zapata.conf',
			     "; Channel $channel START",
			     'signalling = '.$fx.'_ks', # Assume FXS port
			     "context=$context",
			     "channel => $channel",
			     "; Channel $channel END",
			     );
	}
    }
    
    # test to make sure we got the analog channel right:
    if(ztcfg_fails()) {
	my $dchan = 0;
	my $conf_file = '/etc/zaptel.conf';
	my $temp_file = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);
	open(CONF, $conf_file) || warn $!;
	open(TEMP, ">$temp_file") || warn $!;

	while(<CONF>) {
	    if (/^dchan=(\d+)/) {
		$dchan = $1;
		print TEMP "$_";
	    }
	    elsif (/(fx[os]ks=)(\d+)/) {
		my $chan = $2;
		if ($chan < $dchan) {
		    $chan += $dchan;
		    print TEMP "$1$chan\n";
		}
		else {
		    print TEMP "$_";
		}
	    }
	    else {
		print TEMP "$_";
	    }
	}
	close TEMP;
	close CONF;

	copy($temp_file, $conf_file) || warn "$!";

	unlink($temp_file);
    }   
    return 1;
}


##############################################################################
# configure_as_fxs_port: Configure the given port as an FXS port
##############################################################################
sub configure_as_fxs_port {
	my($c) = shift(@_);

	update_conf_file(
		'/etc/zaptel.conf',
		"# Channel $c START",
		"fxoks=$c",
		"# Channel $c END",
	);
	update_conf_file(
		'/etc/asterisk/zapata.conf',
		"; Channel $c START",
		'signalling = fxo_ks', # Assume FXS port
		'context=internal',
		"channel => $c",
		"; Channel $c END",
	);
}

##############################################################################
# configure_as_fxo_port: Configure the given port as an FXO port
##############################################################################
sub configure_as_fxo_port {
	my($c) = shift(@_);

	update_conf_file(
		'/etc/zaptel.conf',
		"# Channel $c START",
		"fxsks=$c",
		"# Channel $c END",
	);
	update_conf_file(
		'/etc/asterisk/zapata.conf',
		"; Channel $c START",
		'signalling = fxs_ks', # FXO port
		'context=incoming',
		'group = 0',
		"channel => $c",
		"; Channel $c END",
	);
}

##############################################################################
# get_chan_details_by_span: Return digital span details for given span num
##############################################################################
sub get_chan_details_by_span {
	my($span_number) = shift(@_);
	my($count) = 0;
	my($first_chan,$last_chan);
	my($first_bchan,$last_bchan,$dchan);

	open(PROC, "/proc/zaptel/$span_number") || warn $!;
	while(<PROC>) {
		if(/^\s+(\d+)/) {
			$first_chan = $1 unless($first_chan);
			$last_chan = $1;
			$count++;
		}
	}
	close(PROC);

	if($count == 24) { # PRI
		$first_bchan = $first_chan;
		$last_bchan = $last_chan - 1;
		$dchan = $last_chan;
	}
	else { # E1 or BRI
		dprint("Unsupported digital span");
		return(undef,undef,undef);
	}

	return($first_bchan,$last_bchan,$dchan);
}

##############################################################################
# get_channel_list_by_span: Return channels hash_ref for a given span number
##############################################################################
sub get_channel_list_by_span {
	my($span_number) = shift(@_);
	my(%channels);

	open(PROC, "/proc/zaptel/$span_number") || warn $!;
	while(<PROC>) {
		if(/^\s+(\d+)\s+(\S.*)$/) {
			$channels{$1} = $2;
		}
	}
	close(PROC);

	return(\%channels);
}
##############################################################################
# prune_old_config: prune the channel parts of old config
##############################################################################

sub prune_old_config {
	my ($conf_file) = @_;
	my ($temp_file) = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);
	
	open(TEMP, ">$temp_file") || warn $!;
	if (! -e $conf_file)
	{
		 force_minimal_zaptel_config();
	}
	open(CONF, $conf_file) || warn $!;
	
	# put everything above empty line into temp
	while(<CONF>) {
		if (/^(\s)*$/){
			print TEMP "\n";
			last;
		}
		else {
			print TEMP "$_";
		}
	}
	close TEMP;

	copy($temp_file, $conf_file);
	unlink($temp_file);
}

##############################################################################
# update_conf_file: Update the given configuration file with given lines
##############################################################################
sub update_conf_file {
	my($conf_file) = shift(@_);
	my(@lines) = @_;
	my($temp_file) = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);
	my($skip) = 0;

	my($first_line) = $lines[0];
	my($last_line) = pop(@lines);
	push(@lines, $last_line);

	open(TEMP, ">$temp_file") || warn $!;
	open(CONF, $conf_file) || warn $!;
	
	while(<CONF>) {
		chomp;
		if(/^$first_line/) {
			$skip++;
			next;
		}
		elsif(/^$last_line/) {
			$skip = 0;
			next;
		}
		if($skip == 0) {
			print TEMP "$_\n";
		}
	}
	close(CONF);
	foreach my $line (@lines) {
		dprint("CONFIG: $conf_file: $line");
		print TEMP "$line\n";
	}
	close(TEMP);
	copy($temp_file,$conf_file) || warn $!;
	unlink($temp_file);
}

##############################################################################
# sync_conf_file: Synchronize configuration file with Fonality's web farm
##############################################################################
sub sync_conf_file {
	my($file) = shift(@_);
	$file =~ s/.conf$//;
	my($rs_cgi) = 'http://cp.trixbox.com/reverse_sync.cgi';
	my($hostname) = Sys::Hostname::hostname();
	if($hostname =~ /^\D+(\d+)$/) {
		my($server_id) = $1;
		my($url) = "${rs_cgi}?account=${server_id}&file=$file";
		dprint("Syncing config file: url=$url");
		LWP::Simple::get($url);
	}
	else {
		vprint("WARNING: Can not determine server_id from hostname");
	}
}

##############################################################################
# prune_conf_entry: Prune the given configuration file
##############################################################################
sub prune_conf_entry {
	my($conf_file) = shift(@_);
	my(@lines) = @_;
	my($temp_file) = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);
	my($skip) = 0;

	my($first_line) = $lines[0];
	my($last_line) = pop(@lines);
	push(@lines, $last_line);

	open(TEMP, ">$temp_file") || warn $!;
	open(CONF, $conf_file) || warn $!;
	while(<CONF>) {
		chomp($_);
		if(/^$first_line/) {
			$skip++;
			next;
		}
		elsif(/^$last_line/) {
			$skip = 0;
			next;
		}
		if($skip == 0) {
			print TEMP "$_\n";
		}
	}
	close(CONF);
	close(TEMP);
	copy($temp_file,$conf_file) || warn $!;
	unlink($temp_file);
}

##############################################################################
# prune_single_line: Prune the given line from the given configuration file
##############################################################################
sub prune_single_line {
	my($conf_file) = shift(@_);
	my($line) = shift(@_);
	my($temp_file) = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);

	open(TEMP, ">$temp_file") || warn $!;
	open(CONF, $conf_file) || warn $!;
	while(<CONF>) {
		chomp($_);
		if(/^$line/) {
			next;
		}
		else {
			print TEMP "$_\n";
		}
	}
	close(CONF);
	close(TEMP);
	copy($temp_file,$conf_file) || warn $!;
	unlink($temp_file);
}

##############################################################################
# irq_problem: Return true if the IRQ's of TigerJet 320 boards overlap
##############################################################################
sub irq_problem {
	my(@driver_watch_list) = qw(wctdm);
	my($problem_found) = 0;
	
	open(IRQ, "cat /proc/interrupts |") || warn $!;
	while(<IRQ>) {
		s/^\s+//;
		if(/(\d+):\s+\d+\s+\S+\s+(.*)/) {
			my($irq) = $1;
			my($data) = $2;
			dprint("IRQ $irq = $data");
			$data =~ s/\s+//g;
			if($data =~ /,/) {
				dprint("IRQ $irq sharing between multiple drivers: $data");
				my(@drivers) = split(/,/, $data);
				foreach my $d (@drivers) {
					foreach my $w (@driver_watch_list) {
						if($d eq $w) {
							vprint("WARNING: Driver '$d' sharing IRQ $irq may cause audio quality problems");
							$problem_found++;
						}
					}
				}
			}
		}
		elsif(/^NMI:\s+(\d+)/) {
			my($nmi) = $1;
			next if($nmi == 0);
			$problem_found++;
			vprint("WARNING: IRQ PROBLEM: NMI = $nmi");
		}
		elsif(/^ERR:\s+(\d+)/) {
			my($err) = $1;
			next if($err == 0);
			$problem_found++;
			vprint("WARNING: IRQ PROBLEM: ERR = $err");
		}
		elsif(/^MIS:\s+(\d+)/) {
			my($mis) = $1;
			next if($mis == 0);
			$problem_found++;
			vprint("WARNING: IRQ PROBLEM: MIS = $mis");
		}
	}
	close(IRQ);

	if($problem_found) {
		return(1);
	}
	else {
		return(0);
	}
}

##############################################################################
# force_minimal_zaptel_config: Make sure zaptel.conf has the minumum needed
##############################################################################
sub force_minimal_zaptel_config {
	my($file) = shift(@_) || '/etc/zaptel.conf';
	my(@lines) = ('loadzone=', 'defaultzone=');
	my($temp_file) = "/tmp/temp.$$." . time() . '.' . $$;
	open(ZAPTEL, $file) || warn $!;
	my(@zaptel_lines) = <ZAPTEL>;
	close(ZAPTEL);
	my(@lines_to_prepend);

	foreach my $line (@lines) { # Scan for required lines
		my($found) = 0;
		foreach my $z_line (@zaptel_lines) {
			if($z_line =~ /^$line/) {
				$found++;
				last;
			}
		}
		if($found == 0) {
			if($line eq 'loadzone=') {
				push(@lines_to_prepend, $line . 'us');
			}
			if($line eq 'defaultzone=') {
				push(@lines_to_prepend, $line . 'us');
			}
		}
	}

	# Always write a default zaptel.conf
	open(TMP, ">$temp_file") || warn $!;
	foreach my $line (@lines_to_prepend) {
		print TMP "$line\n";
	}
	foreach my $line (@zaptel_lines) {
		print TMP $line;
	}
	close(TMP);
	copy($temp_file, $file) && unlink($temp_file);
}

##############################################################################
# ztcfg_fails: Returns true if ztcfg returns a failure
##############################################################################
sub ztcfg_fails {
	my($arg) = shift(@_);

	unless(wait_fot_dev_zap_ctl(10)) {
		vprint("WARNING: Timed out waiting for /dev/zap/ctl");
		return(1);
	}
	if($arg) {
		lsystem("/sbin/ztcfg $arg");
	}
	else {
		 lsystem("/sbin/ztcfg > /dev/null 2>&1");
	}
	return(0) if($? == 0);
	dprint("ztcfg failed");
	return(1);
}

##############################################################################
# wait_fot_dev_zap_ctl: Wait for /dev/zap/ctl to be created
##############################################################################
sub wait_fot_dev_zap_ctl {
	my($max) = shift(@_); # Max seconds to wait
	my($cnt) = 0;

	while(! -e '/dev/zap/ctl') {
		vprint("Waiting for 1 second for drivers to load...");
		return(0) if($cnt >= $max);
		dprint("Waiting for /dev/zap/ctl to appear...");
		sleep(1);
		$cnt++;
	}
	dprint("Found /dev/zap/ctl");

	return(1);
}

##############################################################################
# power_cable_unplugged: Return TRUE if power cable disconnected.
##############################################################################
sub power_cable_unplugged {
	my($log_file) = shift(@_) || '/var/log/messages';
	my(@text_strings) = (
		'DID YOU REMEMBER TO PLUG IN THE HD POWER CABLE TO THE TDM400P',
		'Did you remember to plug in the power cable',
	);
	my($problem_found) = 0;

	open(MSG, $log_file) || warn "Unable to open $log_file because: $!";
	while(<MSG>) {
		foreach my $text (@text_strings) {
			if(/$text/) {
				$problem_found++;
				last;
			}
		}
		last if($problem_found);
	}
	close(MSG);

	if($problem_found && $log_file eq '/var/log/messages') {
		# Rotate the log file so that after boot it will work
		rename($log_file, "$log_file." . time());
		open(MSG, ">$log_file");
		close(MSG);
		open(PID, '/var/run/syslog.pid');
		while(<PID>) {
			if(/^\d+$/) {
				kill(1,$1) && dprint("HUP'd syslogd at PID $1");
			}
		}
		close(PID);
		return(1);
	}
	elsif($problem_found) {
		return(1);
	}

	return(0);
}


##############################################################################
# check_for_chans
##############################################################################
sub check_for_chans {
	my($spans) = shift(@_);
	my($channels) = shift(@_);
	my($new_chans) = 0;
	my($new_spans) = 0;
	my($dir) = '/proc/zaptel';

	my($old_span_count) = scalar(keys(%{$spans}));
	my($new_span_count) = 0;
	my($old_chan_count) = scalar(keys(%{$channels}));
	my($new_chan_count) = 0;

	if(-d $dir) {
		chdir($dir) || warn $!;
		opendir(DIR, $dir) || warn $!;
		my(@files) = sort { $a <=> $b } grep { /^\d+$/ } readdir(DIR);
		closedir(DIR);

		foreach my $f (@files) {
			dprint("Scrubbing proc file $f for info");
			open(PROC, $f) || warn $!;

=pod

Span 1: WCTDM/0 "Wildcard TDM400P REV E/F Board 1"

           1 WCTDM/0/0
           2 WCTDM/0/1
           3 WCTDM/0/2
           4 WCTDM/0/3

=cut

			while(<PROC>) {
				chomp($_);
				if(/^Span (\d+): (\S.*)$/) {
					my($num) = $1;
					my($text) = $2;
					$text =~ s/[\(\)\"]//g;
					$spans->{$num} = $text;
					$new_span_count++;
				}
				elsif(/^\s+(\d+)\s+(\S+)/) {
					$channels->{$1} = $2;
					$new_chan_count++;
				}
				dprint($_);
			}
			close(PROC);
		}
	}
	else {
		dprint("check_for_spans(): '$dir' directory is missing!");
	}

	$new_chans = $new_chan_count - $old_chan_count;
	$new_spans = $new_span_count - $old_span_count;
	
	return($spans,$channels,$new_chans,$new_spans);
}

##############################################################################
# load_module: Load the given module into the kernel
##############################################################################
sub load_module {
	my($name) = shift(@_);

	return(1) if(module_is_loaded($name)); # Don't attempt to load twice

	vprint("Loading module: $name");

	if($name eq 'wanpipe') { 
		start_wanrouter();
		return(1);
	}

	lsystem("/sbin/modprobe $name");
	unless(wait_fot_dev_zap_ctl(20)) {
		vprint("WARNING: Timed out waiting for /dev/zap/ctl");
		return(0);
	}
	die "Unable to load module '$name' because: $!" unless($? == 0);
	if(irq_problem()) {
		vprint("IRQ problem detected");	
	}
	return(1);
}

##############################################################################
# unload_module: Unload the given module from the kernel
##############################################################################
sub unload_module {
	my($name) = shift(@_);

	return(1) unless(module_is_loaded($name)); # Only remove if exists

	vprint("Unloading module: $name");
	lsystem("/sbin/modprobe -r $name > /dev/null 2>&1");
	unless($? == 0) {
		#my(@text) = `/sbin/lsmod | grep zaptel`;
		#dprint("@text") if(scalar(@text));
		vprint("Unable to unload module '$name' because: $!");
		return(2);
	}
	return(0);
}

##############################################################################
# unload_all_modules: Unload all the zaptel modules form the kernel
##############################################################################
sub unload_all_modules {
	my(@names) = qw(
		rcbfx
		wctdm
		r4fxo
		wcfxo
		wcfxs
		pciradio
		tor2
		torisa
		r1t1
		wct1xxp
		rxt1
		wct4xxp
		wcte11xp
		wcusb
		ztd_eth
		ztdummy
		zttranscode
	);

	stop_asterisk();
	stop_wanrouter();

	foreach my $name (@names) {
		unload_module($name);
	}

	open(LSMOD, "/sbin/lsmod | ") || warn $!;
	while(<LSMOD>) {
		#zaptel                158628  2 ztdummy,wctdm
		if(/^zaptel\s+\d+\s+\d+\s+(\S+)/) {
			my($csv) = $1;
			if($csv =~ /,/) {
				foreach my $m (split(/,/, $csv)) {
					unload_module($m);
				}
			}
			else {
				unload_module($csv);
			}
		}
	}
	close(LSMOD);

	for my $i (1..10) {
		if(module_is_loaded('zaptel')) {
			unload_module('zaptel');
		}
		if(module_is_loaded('zaptel')) {
			dprint("Module 'zaptel' is still loaded, sleeping and retrying");
			sleep(1);
			unload_module('zaptel');
		}
		else {
			last;
		}
	}

	if(module_is_loaded('zaptel')) {
		die "Unable to unload zaptel module(s). Manual intervention required.";
	}

	return(1);
}

##############################################################################
# module_is_loaded: Return TRUE if the given module name is loaded
##############################################################################
sub module_is_loaded {
	my($name) = shift(@_);
	my($module_found) = 0;

	open(LSMOD, "/sbin/lsmod | ") || warn $!;
	while(<LSMOD>) {
		if(/$name/) {
			$module_found++;
			last;
		}
	}
	close(LSMOD);
	
	return($module_found);
}

##############################################################################
# asterisk_is_running: Returns TRUE if asterisk is running
##############################################################################
sub asterisk_is_running {
	my($pid) = '/var/run/asterisk.pid';
	my($ctl) = '/var/run/asterisk.ctl';
	if(-f $pid) {
		dprint("Found PID file: $pid");
		if(-S $ctl) {
			open(PID, $pid) || warn $!;
			my(@lines) = <PID>;
			close(PID);
			chomp(@lines);
			if($lines[0] =~ /^\d+$/) {
				dprint("Checking for PID $lines[0]");
				if(kill(0,$lines[0])) { # Running
					dprint("PID $lines[0] is running");
					return(1);
				}
			}
		}
	}
	return(0);
}

##############################################################################
# stop_asterisk: Stop asterisk quickly
##############################################################################
sub stop_asterisk {

	# Stop astwatch first
	lsystem("/usr/bin/killall -9 astwatch.pl > /dev/null 2>&1");

	if(asterisk_is_running()) {
		vprint("Stopping PBXtra Core");
		dprint("Sending SIGKILL to PBXtra Core");
		lsystem("/usr/sbin/asterisk -rx 'stop now'");
	}
	else {
		dprint("PBXtra Core already stopped. Not trying to stop it.");
		return(0);
	}

	for my $i (1..10) {
		if(asterisk_is_running()) {
			vprint("Stopping PBXtra Core");
			dprint("Sending SIGKILL to PBXtra Core");
			lsystem("/usr/bin/killall -9 /usr/sbin/asterisk > /dev/null 2>&1");
		}
		else {
			last;
		}
	}
	for my $i (1..10) {
		if(asterisk_is_running()) {
			vprint("Stopping PBXtra Core");
			dprint("Sending SIGKILL to PBXtra Core");
			lsystem("/usr/bin/killall -9 /usr/sbin/asterisk > /dev/null 2>&1");
			sleep(1);
		}
		else {
			last;
		}
	}
	if(asterisk_is_running()) {
		die "Unable to stop PBXtra Core. Manual intervention required.";
	}
	return(1);
}

##############################################################################
# show_supported_cards: Show a list of supported cards
##############################################################################
sub show_supported_cards {
	my($cards) = get_card_info();
	my($cnt_variants) = 0;
	my($cnt_cards) = 0;
	my(%c);

	print "Supported Cards:\n\n";
	foreach my $key (keys(%{$cards})) {
		my($desc) = $cards->{$key}->{'description'};
		$c{$desc}++;
		$cnt_variants++;
	}
	foreach my $key (sort(keys(%c))) {
		print $key;
		if($c{$key} > 1) {
			print " [$c{$key} variants]\n";
		}
		else {
			print "\n";
		}
		$cnt_cards++;
	}

	print "\nSummary: $cnt_variants total variants of $cnt_cards cards.\n";
}

##############################################################################
# get_card_info:
##############################################################################
sub get_card_info {
	my(%card_info) = (
		'Subsystem: a010:' => {
			'description' => 'Sangoma A101 with Single T1/E1 line',
			'manufacturer' => 'Sangoma',
			'model' => 'A101',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a011:' => {
			'description' => 'Sangoma A102 with Dual T1/E1',
			'manufacturer' => 'Sangoma',
			'model' => 'A102',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a013:' => {
			'description' => 'Sangoma A104 with Quad T1/E1 lines',
			'manufacturer' => 'Sangoma',
			'model' => 'A101',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a030:' => {
			'description' => 'Sangoma A305 with Channelized T3',
			'manufacturer' => 'Sangoma',
			'model' => 'A305',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a111:' => {
			'description' => 'Sangoma A101 Shark with Single T1/E1',
			'manufacturer' => 'Sangoma',
			'model' => 'A101',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a112:' => {
			'description' => 'Sangoma A102 Shark with Dual T1/E1',
			'manufacturer' => 'Sangoma',
			'model' => 'A102',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a113:' => {
			'description' => 'Sangoma A104 Shark with Quad T1/E1',
			'manufacturer' => 'Sangoma',
			'model' => 'A104',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a114:' => {
			'description' => 'Sangoma A108 Shark with Octal T1/E1',
			'manufacturer' => 'Sangoma',
			'model' => 'A108',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a116:' => {
			'description' => 'Sangoma A305 Shark with Channelized T3',
			'manufacturer' => 'Sangoma',
			'model' => 'A305',
			'type' => 'digital',
			'driver' => 'wanpipe',
		},
		'Subsystem: a200:' => {
			'description' => 'Sangoma A200 AFT-REMORA Shark analog board',
			'manufacturer' => 'Sangoma',
			'model' => 'A200',
			'type' => 'analog',
			'driver' => 'wanpipe',
		},
		'Subsystem: a400:' => {
			'description' => 'Sangoma A400 AFT-REMORA Shark analog board',
			'manufacturer' => 'Sangoma',
			'model' => 'A400',
			'type' => 'analog',
			'driver' => 'wanpipe',
		},
		'Subsystem: a500:' => {
			'description' => 'Sangoma A500 ISDN-BRI Shark board',
			'manufacturer' => 'Sangoma',
			'model' => 'A500',
			'type' => 'smg',
			'driver' => 'wanpipe',
		},
		'Class 0780: d161:0205' => {
			'description' => 'Digium Wildcard TE205P',
			'manufacturer' => 'Digium',
			'model' => 'TE205P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: d161:0210' => {
			'description' => 'Digium Wildcard TE210P',
			'manufacturer' => 'Digium',
			'model' => 'TE210P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: d161:0405' => {
			'description' => 'Digium Wildcard TE405P Quad-Span togglable E1/T1/J1 card 5.0v',
			'manufacturer' => 'Digium',
			'model' => 'TE405P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: d161:0406' => {
			'description' => 'Digium Wildcard TE406P Quad-Span togglable E1/T1/J1 echo cancellation card 5.0v',
			'manufacturer' => 'Digium',
			'model' => 'TE406P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: d161:0410' => {
			'description' => 'Digium Wildcard TE410P Quad-Span togglable E1/T1/J1 card 3.3v',
			'manufacturer' => 'Digium',
			'model' => 'TE410P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: d161:0411' => {
			'description' => 'Digium Wildcard TE411P Quad-Span togglable E1/T1/J1 echo cancellation card 3.3v',
			'manufacturer' => 'Digium',
			'model' => 'TE411P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: d161:2400' => {
			'description' => 'Digium Wildcard TDM2400P',
			'manufacturer' => 'Digium',
			'model' => 'TDM2400',
			'type' => 'analog',
			'driver' => 'wctdm24xxp',
		},
		'Class 0780: 10ee:0205' => {
			'description' => 'Digium/Xilinx Wildcard TE205P',
			'manufacturer' => 'Digium',
			'model' => 'TE205P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: 10ee:0210' => {
			'description' => 'Digium/Xilinx Wildcard TE210P',
			'manufacturer' => 'Digium',
			'model' => 'TE210P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: 10ee:0314' => {
			'description' => 'Digium/Xilinx Wildcard TE405P/TE410P (1st Gen)',
			'manufacturer' => 'Digium',
			'model' => 'TE410P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: 10ee:0405' => {
			'description' => 'Digium/Xilinx Wildcard TE405P (2nd Gen)',
			'manufacturer' => 'Digium',
			'model' => 'TE405P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0780: 10ee:0410' => {
			'description' => 'Digium/Xilinx Wildcard TE410P (2nd Gen)',
			'manufacturer' => 'Digium',
			'model' => 'TE410P',
			'type' => 'digital',
			'driver' => 'wct4xxp',
		},
		'Class 0200: d161:8005' => {
			'description' => 'Digium/Xilinx Wildcard TDM410P',
			'manufacturer' => 'Digium',
			'model' => 'TDM410P',
			'type' => 'analog',
			'driver' => 'wctdm24xxp',
		},
		'Class 0280: e159:0001' => {
			'description' => 'Digium/TigerJet 3XX Analog X100P/X101P/TDM400P/A400P',
			'manufacturer' => 'Digium',
			'model' => 'TDM400P',
			'type' => 'analog',
			'driver' => 'wctdm',
		},
		'Class 0780: e159:0001' => {
			'description' => 'Digium/TigerJet 3XX Analog X100P/X101P/TDM400P/A400P',
			'manufacturer' => 'Digium',
			'model' => 'TDM400P',
			'type' => 'analog',
			'driver' => 'wctdm',
		},
		'Class 0200: d161:0800' => {
			'description' => 'Digium TDM808',
			'manufacturer' => 'Digium',
			'model' => 'TDM808B',
			'type' => 'analog',
			'driver' => 'wctdm',
		},
		'Class 0200: d161:0120' => {
			'description' => 'Digium TE120',
			'manufacturer' => 'Digium',
			'model' => 'TE120P',
			'type' => 'digital',
			'driver' => 'wcte12xp',
		},
		'Class 0280: 1057:5608' => {
			'description' => 'Motorola Wildcard X100P',
			'manufacturer' => 'Motorola',
			'model' => 'X100P',
			'type' => 'analog',
			'driver' => 'wctdm',
		},
		'Class 0780: 1057:5608' => {
			'description' => 'Motorola Wildcard X100P',
			'manufacturer' => 'Motorola',
			'model' => 'X100P',
			'type' => 'analog',
			'driver' => 'wcfxo',
		},
		'Class 0280: 0b0b:0105' => {
			'description' => 'Rhino R1T1',
			'manufacturer' => 'Rhino',
			'model' => 'R1T1',
			'type' => 'digital',
			'driver' => 'r1t1',
		},
		'Class 0280: 0b0b:0205' => {
			'description' => 'Rhino R4FXO',
			'manufacturer' => 'Rhino',
			'model' => 'R4FXO',
			'type' => 'analog',
			'driver' => 'rcbfx',
		},
		'Class 0280: 0b0b:0305' => {
			'description' => 'Rhino R4T1',
			'manufacturer' => 'Rhino',
			'model' => 'R4T1',
			'type' => 'digital',
			'driver' => 'rxt1',
		},
		'Class 0280: 0b0b:0405' => {
			'description' => 'Rhino R8FXX',
			'manufacturer' => 'Rhino',
			'model' => 'R8FXX',
			'type' => 'analog',
			'driver' => 'rcbfx',
		},
		'Class 0280: 0b0b:0505' => {
			'description' => 'Rhino R24FXX',
			'manufacturer' => 'Rhino',
			'model' => 'R24FXX',
			'type' => 'analog',
			'driver' => 'rcbfx',
		},
		'Class 0280: 0b0b:0506' => {
			'description' => 'Rhino R2T1',
			'manufacturer' => 'Rhino',
			'model' => 'R2T1',
			'type' => 'digital',
			'driver' => 'rxt1',
		},
		'Class 0280: 0b0b:0605' => {
			'description' => 'Rhino R2T1',
			'manufacturer' => 'Rhino',
			'model' => 'R2T1',
			'type' => 'digital',
			'driver' => 'rxt1',
		},
		'Class 0280: 0b0b:0705' => {
			'description' => 'Rhino R24FXS',
			'manufacturer' => 'Rhino',
			'model' => 'R1T1',
			'type' => 'analog',
			'driver' => 'rcbfx',
		},
		'Class 0200: d161:0800' => {
		    'description' => 'Digium Wildcard TDM800P',
		    'manufacturer' => 'Digium',
		    'model' => 'TDM800P',
		    'type' => 'analog',
		    'driver' => 'wctdm24xxp',
		},
		'Class 0280: 0b0b:0206' => {
		    'description' => 'Rhino R4FXO-EC',
		    'manufacturer' => 'Rhino',
		    'model' => 'R4FXO',
		    'type' => 'analog',
		    'driver' => 'rcbfx',
		}
	);
	dprint("Loaded " . scalar(keys(%card_info)) . " card specifications");
	return(\%card_info);
}

##############################################################################
# dprint: Print output if DEBUG is enabled
##############################################################################
sub dprint {
	return(0) unless($DEBUG);
	print "DEBUG: @_\n";
}

##############################################################################
# vprint: Print output if VERBOSE is enabled
##############################################################################
sub vprint {
	warn "@_\n" if("@_" =~ /^WARNING/);
	return(0) unless($VERBOSE || $DEBUG);
	print "@_\n";
}

##############################################################################
# start_wanrouter: Start Sangoma wanrouter/wanpipe
##############################################################################
sub start_wanrouter {
	my($wr) = '/usr/sbin/wanrouter';
	return(undef) unless(-x $wr);
	lsystem("$wr start > /dev/null 2>&1");
	dprint("Non-zero exit status from 'wanrouter start': $?") if($? != 0);
	return($?);
}

##############################################################################
# stop_wanrouter: Stop Sangoma wanrouter/wanpipe
##############################################################################
sub stop_wanrouter {
	my($wr) = '/usr/sbin/wanrouter';
	return(undef) unless((-x $wr) && module_is_loaded('wanpipe'));
	lsystem("$wr stop > /dev/null 2>&1");
	dprint("Non-zero exit status from 'wanrouter stop': $?") if($? != 0);
	return($?);
}

##############################################################################
# get_port_details: Query zaptel cards for port type details
##############################################################################
sub get_port_details {
        my ($card) = @_;

	my $ZT_GET_PARAMS = -2138813947;
	my $__ZT_SIG_FXO = 4096; # Hex = 1000
	my $__ZT_SIG_FXS = 8192; # Hex = 2000

	wait_fot_dev_zap_ctl(20) || return(undef);

	open(CTL, "+</dev/zap/ctl") || return "Couldn't open /dev/zap/ctl for rw: $!";

	my @cards;
	my @chans = $card ? ($card) : (1 .. 12);
	foreach my $chan (@chans)
	{
	        my $zt_params = zt_params({ channo => $chan });
	        ioctl(CTL, $ZT_GET_PARAMS, $zt_params) || last;
		my(%h) = zt_params_unp($zt_params);

		my($ztp_sigcap) = $h{'sigcap'};
		my($val) = $ztp_sigcap & ($__ZT_SIG_FXO|$__ZT_SIG_FXS);
		if(($val & $__ZT_SIG_FXO) && ($val & $__ZT_SIG_FXS)) {
			$h{'porttype'} = 'NONE';
		}
		elsif($val & $__ZT_SIG_FXS) {
			$h{'porttype'} = 'FXO';
		}
		elsif($val & $__ZT_SIG_FXO) {
			$h{'porttype'} = 'FXS';
		}
	        push(@cards, \%h);
	}
	close(CTL);

	return \@cards;
}

##############################################################################
# zt_params: return a packed scructure to pass to C
##############################################################################
sub zt_params {
	my(%tmp) = %{$_[0]};

	return
	pack("iiiiiiiiiiiiA40iiiiiiiiiii",
		$tmp{channo},
		$tmp{spanno},
		$tmp{chanpos},
		$tmp{sigtype},
		$tmp{sigcap},
		$tmp{rxisoffhook},
		$tmp{rxbits},
		$tmp{txbits},
		$tmp{txhooksig},
		$tmp{rxhooksig},
		$tmp{curlaw},
		$tmp{idlebits},
		$tmp{name},
		$tmp{prewinktime},
		$tmp{preflashtime},
		$tmp{winktime},
		$tmp{flashtime},
		$tmp{starttime},
		$tmp{rxwinktime},
		$tmp{rxflashtime},
		$tmp{debouncetime},
		$tmp{pulsebreaktime},
		$tmp{pulsemaketime},
		$tmp{pulseaftertime},
	);
}

##############################################################################
# zt_params_unp: Unpack a packed C scritcure and return it as a hash
##############################################################################
sub zt_params_unp {
	my($tmp) = shift(@_);
	my(@tmp) = unpack("iiiiiiiiiiiiA40iiiiiiiiiii", $tmp);

	my($i) = 0;
	return map { $_ => $tmp[$i++] }
	qw(
		channo
		spanno
		chanpos
		sigtype
		sigcap
		rxisoffhook
		rxbits
		txbits
		txhooksig
		rxhooksig
		curlaw
		idlebits
		name
		prewinktime
		preflashtime
		winktime
		flashtime
		starttime
		rxwinktime
		rxflashtime
		debouncetime
		pulsebreaktime
		pulsemaketime
		pulseaftertime
	);
}

##############################################################################
# zaptel_version_supports_autoconfig: Return TRUE if we have the right zaptel
##############################################################################
sub zaptel_version_supports_autoconfig {
	my($required_version) = 'zaptel-1.2.16-5';
	lsystem("/bin/rpm -q $required_version > /dev/null");
	return(1) if($? == 0);
	return(0);
}
sub wanrouter_supports_silent {
	my($required_version) = '3.3.6';
	my($version) = `rpm -q --queryformat %{VERSION} wanpipe-util`;
	if ($version >= $required_version)
	{
		# Latest version of wanpipe-util
		# support for silent option
		return 1;
	}
	else
	{
		# Old version of wanpipe-util
		return 0;
	}
}
sub configure_woomera_module {
	my($bri_enable_module);
	if (-e "/etc/wanpipe/smg_bri.conf") {
		open BRI, "/etc/wanpipe/smg_bri.conf" || warn $!;
		while (<BRI>) {   
	        	if (/^spans.*/) {   
        			$bri_enable_module = 1 ;#enable module
            			last;
        		}
		}
	close BRI;	
	}
   	my($modconf_file) = "/etc/asterisk/modules.conf";
   	my($temp_file) = '/tmp/temp.' . time() . '.' . $$ . '.' . rand(9999);

    	open TEMP, ">$temp_file" || warn $!;
    	open MOD, $modconf_file || warn $!;
    	while(<MOD>) {  
		#check for the "noload" entry first
        	if (/noload.*chan_woomera.so/) {
        		if ($bri_enable_module) {
				next;
			}
        	}
        	elsif (/\[global\]/) {
			unless($bri_enable_module) {
				print TEMP "noload => chan_woomera.so\n";
			}
        		print TEMP "$_";
        	}
		#check for chan_woomera.so=no
		elsif (/chan_woomera.so=no/) {
                       if ($bri_enable_module) {
				next;
                       }
                } 
		elsif (eof(MOD)) {
			if ($bri_enable_module) {
				print TEMP "$_";
                        } else { 
                                print TEMP "chan_woomera.so=no\n";
                                print TEMP "$_";
                        }
		} else { print TEMP "$_"; }
    	}
    	close TEMP;
    	close OUT;
	close MOD;

    	copy($temp_file, $modconf_file) || warn $!;
	unlink($temp_file);
}


sub config_wanpipe { 

	my($WANONLY) = shift();
	my($OPTIONS) = "--trixbox";

	if (wanrouter_supports_silent()) {
		## 3.3.6 needs --no_boot otherwise sangoma activates boot scripts by default :P	
		substr($OPTIONS, 0, 0, "--silent --no_boot ");
	}

	if ($WANONLY) {
		$OPTIONS = "--silent --no_boot --no-zaptel --no-zapata";
	}
	
	lsystem("/usr/local/sbin/setup-sangoma-fon $OPTIONS > /dev/null 2>&1");
}

sub lsystem
{
#	print STDERR "system(\"@_\")\n";
	system(@_);
}
